// mnyfmt.h (C) 2011 adolfo@di-mare.com
// mnyfmt.c: A simple function to format money or currency amounts

/** \file   mnyfmt.h
    \brief  Header file for \c mnyfmt().

    \author Adolfo Di Mare <adolfo@di-mare.com>
    \date   2011
*/

#ifndef MNYFMT_ADOLFO
#define MNYFMT_ADOLFO

#ifndef MNYFMT_NO_LONG_LONG
    typedef long long mnyfmt_long;
#else
    typedef long mnyfmt_long;
#endif

/// Formatting character for \c mnyfmt().
#define mnyfmt_format_char (char)('9')

#ifdef __cplusplus
extern "C" {
#endif

// Examples and unit test cases are in in file mnyfmtts.c
// The complete Doxygen documentation is in file mnyfmt.c
char* mnyfmt(char *fmtstr, char dec, mnyfmt_long intpart, unsigned fractpart);

// Version that uses the number separated in integer and fractiona part
// - To compile, include mnyfmtdd.c in your project.
char* mnyfmtdd(char *fmtstr, char dec, double intdouble, double fractdouble);

#ifdef __cplusplus
}
#endif

#ifdef __cplusplus
    // shortcut for integer only formatting
    inline char* mnyfmt0(char *fmtstr, mnyfmt_long intpart ) {
        return mnyfmt(fmtstr, '\0', intpart, 0 );
    }
#else
    #if 0
        // Does C have the "inline" optimization?
        static char* mnyfmt0(char *fmtstr, mnyfmt_long intpart ) {
            return mnyfmt(fmtstr, '\0', intpart, 0 ) {
        }
    #else
        // Use a macro to inline the invocation.
        #define mnyfmt0(fmtstr,intpart) mnyfmt(fmtstr, '\0', intpart, 0 )
    #endif
#endif

/// \fn    mnyfmt0(char*,mnyfmt_long)
/// \brief <code>return mnyfmt(fmtstr, '\0', intpart, 0 )</code>

#endif
// EOF: mnyfmt.h
